#include <ntifs.h>
#include "fastio.h"
#include "hash.h"
#include "F_IFace.h"
#include "main.h"
#include "filehiddenroutine.h"
#include "utility.h"
#include "Recognizer.h"

PDRIVER_OBJECT g_DriverObject=NULL;
PDEVICE_OBJECT g_DeviceObject=NULL;
HANDLE g_EventHandle=NULL;
PKEVENT g_pEventObject=NULL;

LIST_ENTRY FsDeviceQueue;
ERESOURCE FsLock;
STORENAMEINSTANCE DriveNameList[MAXDRIVENUMBER];

extern ERESOURCE LockHash;
extern PHASH_ENTRY FreeHashList;
extern PHASH_ENTRY HashTable[NUM_HASH];

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, DriverEntry)
#pragma alloc_text(PAGE, DriverUnload)
#pragma alloc_text(PAGE, FILEHIDECreate)
#pragma alloc_text(PAGE, FILEHIDEFastIoCheckIfPossible)
#pragma alloc_text(PAGE, FILEHIDEFastIoRead)
#pragma alloc_text(PAGE, FILEHIDEFastIoWrite)
#pragma alloc_text(PAGE, FILEHIDEFastIoQueryBasicInfo)
#pragma alloc_text(PAGE, FILEHIDEFastIoQueryStandardInfo)
#pragma alloc_text(PAGE, FILEHIDEFastIoLock)
#pragma alloc_text(PAGE, FILEHIDEFastIoUnlockSingle)
#pragma alloc_text(PAGE, FILEHIDEFastIoUnlockAll)
#pragma alloc_text(PAGE, FILEHIDEFastIoUnlockAllByKey)
#pragma alloc_text(PAGE, FILEHIDEFastIoDeviceControl)
#pragma alloc_text(PAGE, FILEHIDEFastIoDetachDevice)
#pragma alloc_text(PAGE, FILEHIDEFastIoQueryNetworkOpenInfo)
#pragma alloc_text(PAGE, FILEHIDEFastIoMdlRead)
#pragma alloc_text(PAGE, FILEHIDEFastIoPrepareMdlWrite)
#pragma alloc_text(PAGE, FILEHIDEFastIoMdlWriteComplete)
#pragma alloc_text(PAGE, FILEHIDEFastIoReadCompressed)
#pragma alloc_text(PAGE, FILEHIDEFastIoWriteCompressed)
#pragma alloc_text(PAGE, FILEHIDEFastIoQueryOpen)
#endif

NTSTATUS SkipAndSendIrpToNextDrive(IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp)
{
	PDEVICE_EXTENSION pDevExt;
	pDevExt = pDevObj->DeviceExtension;

	IoSkipCurrentIrpStackLocation( pIrp );
	return IoCallDriver(pDevExt->NextDriverDeviceObject, pIrp);
}

NTSTATUS ReturnCurrentStackAfterSendIrpToNextDrive(IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp)
{
	PDEVICE_EXTENSION pDevExt;
	KEVENT Event;
	NTSTATUS ntStatus;
	
	pDevExt = pDevObj->DeviceExtension;

	IoCopyCurrentIrpStackLocationToNext( pIrp );

	KeInitializeEvent( &Event, NotificationEvent, FALSE );

	IoCopyCurrentIrpStackLocationToNext( pIrp );
	IoSetCompletionRoutine(
					pIrp,
					FILEHIDECommonCompletion,
					&Event,
					TRUE,
					TRUE,
					TRUE
					);

	ntStatus = IoCallDriver( pDevExt->NextDriverDeviceObject, pIrp );
	if( ntStatus == STATUS_PENDING )
	{
		KeWaitForSingleObject( &Event, Executive, KernelMode, FALSE, 0 );
		ntStatus = pIrp->IoStatus.Status;
	}
	return ntStatus;
}

NTSTATUS FILEHIDEDeviceIoControl(IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp)
{
	PIO_STACK_LOCATION pCurIrpStack;
	PDEVICE_EXTENSION pDevExt;
	NTSTATUS ntStatus;
	ULONG ulIoControlCode;
	PVOID pInputBuffer, pOutputBuffer;
	ULONG ulInputBufferLength, ulOutputBufferLength;
	
	if( g_DeviceObject != pDevObj )
	{
		return SkipAndSendIrpToNextDrive( pDevObj, pIrp );
	}
		
	pDevExt = pDevObj->DeviceExtension;
	pIrp->IoStatus.Information = 0;
	pCurIrpStack = IoGetCurrentIrpStackLocation(pIrp);
   	ulIoControlCode = pCurIrpStack->Parameters.DeviceIoControl.IoControlCode;
    
    	pInputBuffer = pIrp->AssociatedIrp.SystemBuffer;
	pOutputBuffer = pIrp->AssociatedIrp.SystemBuffer;
    	ulInputBufferLength = pCurIrpStack->Parameters.DeviceIoControl.InputBufferLength;
    	ulOutputBufferLength = pCurIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
	
	switch(ulIoControlCode) 
	{
		case IOCTL_SET_EVENT:
			{
				PSEND_EVENT pEvent;
				pEvent = (PSEND_EVENT)pInputBuffer;

			    	ntStatus = ObReferenceObjectByHandle( pEvent->hEvent,
                                        0x80000000,
                                        NULL,
                                        0,
                                        &g_pEventObject,
                                        NULL );
				if( ntStatus == STATUS_SUCCESS )
				{
				       KeClearEvent(g_pEventObject);
					g_EventHandle = pEvent->hEvent;
				}
				else
				{
					g_pEventObject = NULL;
					g_EventHandle = NULL;
				}
				pIrp->IoStatus.Information = 0;
			}
			break;

		case IOCTL_SET_DRIVE:
			{
				PSEND_DRIVE pDrive;
				pDrive = (PSEND_DRIVE)pInputBuffer;
				SetSymbolicDriveName(pDrive->Drive); 	// 'A' or 'a' ..
				ntStatus = STATUS_SUCCESS;
				pIrp->IoStatus.Information = 0;
			}
			break;

#ifdef _FILEHIDDEN_H
		case IOCTL_HIDDEN_FILES:
		case IOCTL_SHOW_FILES:
		case IOCTL_ADD_HIDDEN_FILE:
		case IOCTL_DEL_HIDDEN_FILE:
		case IOCTL_ALL_RESET_HIDDEN_FILES:
			return HiddenFilesRoutine( pDevObj, pIrp );

#endif //_FILEHIDDEN_H

		default:
			ntStatus = STATUS_SUCCESS;
			pIrp->IoStatus.Information = 0;
			break;
	}

	pIrp->IoStatus.Status = ntStatus;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	return ntStatus;
}

#ifdef _FILEHIDDEN_H

NTSTATUS
FILEHIDEDirectoryControl(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp
	)
{
	PDEVICE_EXTENSION deviceExtension;
	NTSTATUS ntStatus;
	PIO_STACK_LOCATION irpSp = IoGetCurrentIrpStackLocation( Irp );
	PFILE_BOTH_DIR_INFORMATION pInfo;
	ULONG dwInformation;
	PFILE_OBJECT FileObj;
	PWCHAR pwFileName;
	ULONG dwFileDirecotryInfoSize;

	FileObj = irpSp->FileObject;

	deviceExtension = DeviceObject->DeviceExtension;

	if( irpSp -> MinorFunction != IRP_MN_QUERY_DIRECTORY  )
	{
		return SkipAndSendIrpToNextDrive( DeviceObject, Irp );
	}

	if( irpSp -> Parameters.QueryDirectory.FileInformationClass  != FileBothDirectoryInformation  )
	{
		return SkipAndSendIrpToNextDrive( DeviceObject, Irp );
	}

	if( !fIsExistInHashTable( FileObj, &pwFileName ) )
	{
		return SkipAndSendIrpToNextDrive( DeviceObject, Irp );
	}

	if( !pwFileName )
	{
		return SkipAndSendIrpToNextDrive( DeviceObject, Irp );
	}
	
	if( Irp->AssociatedIrp.SystemBuffer )
		pInfo = Irp->AssociatedIrp.SystemBuffer;
	else 
		pInfo = Irp->UserBuffer;

	ntStatus = ReturnCurrentStackAfterSendIrpToNextDrive(DeviceObject, Irp);
	
	if( ntStatus != STATUS_SUCCESS )
	{
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return ntStatus;
	}
	
	dwFileDirecotryInfoSize = GetRealSize( pInfo );
	if( !dwFileDirecotryInfoSize )
	{
		IoCompleteRequest( Irp, IO_NO_INCREMENT );
		return ntStatus;
	}

	dwInformation = HideFiles( pwFileName, pInfo, dwFileDirecotryInfoSize );
	
	Irp->IoStatus.Information = dwInformation;
	IoCompleteRequest( Irp, IO_NO_INCREMENT );
	return ntStatus;
}
#endif //_FILEHIDDEN_H

NTSTATUS
DriverUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
	UNICODE_STRING uniWin32NameString;
	PDEVICE_OBJECT DeviceObject, fsDevice, nextAttachedDevice;
	BOOLEAN bRet=TRUE;
	PDEVICE_EXTENSION deviceExtension;

	FsRtlEnterFileSystem();
	ExAcquireResourceExclusive( &FsLock, TRUE );

	DeviceObject = DriverObject->DeviceObject;

	fsDevice = DeviceObject;

	while(fsDevice->NextDevice)
	{
		nextAttachedDevice = fsDevice->AttachedDevice;
		if( nextAttachedDevice )
		{
			bRet=FALSE;
			break;
		}
		fsDevice = fsDevice->NextDevice;
	}

	if( bRet == FALSE )
		return FALSE; // ̹ ų ϴ
	
	fsDevice = DeviceObject;

	while(fsDevice->NextDevice)
	{
		DeviceObject = fsDevice->NextDevice;
		deviceExtension = (PDEVICE_EXTENSION)fsDevice->DeviceExtension;
		IoDetachDevice( deviceExtension->NextDriverDeviceObject );
		IoDeleteDevice( fsDevice );
		fsDevice = DeviceObject;
	}
		
	IoUnregisterFsRegistrationChange( g_DriverObject, NULL );

	ExFreePool (DriverObject->FastIoDispatch);

	ExReleaseResource( &FsLock );
	FsRtlExitFileSystem();

	ExDeleteResource( &FsLock );

	RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );
	IoDeleteSymbolicLink( &uniWin32NameString );

	IoDeleteDevice( DriverObject->DeviceObject );

	return TRUE;
}

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
	UNICODE_STRING nameString;
	UNICODE_STRING uniWin32NameString;
	PFILE_OBJECT fileObject;
	NTSTATUS status;
	PFAST_IO_DISPATCH fastIoDispatch;
	ULONG i, t;
	PDEVICE_EXTENSION deviceExtension;

#ifdef _FILEHIDDEN_H
	status = InitializeHiddenFileList();
	if( status != STATUS_SUCCESS )
		return status;
#endif //_FILEHIDDEN_H

	ResetList();

	RtlInitUnicodeString( &nameString, L"\\Device\\filehide" );
	status = IoCreateDevice(
	            DriverObject,
	            sizeof( DEVICE_EXTENSION ),
	            &nameString,
	            FILE_DEVICE_DISK_FILE_SYSTEM,
	            0,
	            FALSE,
	            &g_DeviceObject
	            );
	
	if (!NT_SUCCESS( status )) 
	{
	    return status;
	}
		
	DriverObject->DriverUnload = DriverUnload;

	for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++) 
	{
	    DriverObject->MajorFunction[i] = FILEHIDEPassThrough;
	}
	
	DriverObject->MajorFunction[IRP_MJ_CREATE] = FILEHIDECreate;
	DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] = FILEHIDEFsControl;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = FILEHIDEDeviceIoControl;
#ifdef _FILEHIDDEN_H	
	DriverObject->MajorFunction[IRP_MJ_DIRECTORY_CONTROL] = FILEHIDEDirectoryControl;
#endif //_FILEHIDDEN_H    

	fastIoDispatch = ExAllocatePool( NonPagedPool, sizeof( FAST_IO_DISPATCH ) );
	if (!fastIoDispatch) {
	    IoDeleteDevice( g_DeviceObject );
	    return STATUS_INSUFFICIENT_RESOURCES;
	    }

	RtlZeroMemory( fastIoDispatch, sizeof( FAST_IO_DISPATCH ) );
	fastIoDispatch->SizeOfFastIoDispatch = sizeof( FAST_IO_DISPATCH );
	fastIoDispatch->FastIoCheckIfPossible = FILEHIDEFastIoCheckIfPossible;
	fastIoDispatch->FastIoRead = FILEHIDEFastIoRead;
	fastIoDispatch->FastIoWrite = FILEHIDEFastIoWrite;
	fastIoDispatch->FastIoQueryBasicInfo = FILEHIDEFastIoQueryBasicInfo;
	fastIoDispatch->FastIoQueryStandardInfo = FILEHIDEFastIoQueryStandardInfo;
	fastIoDispatch->FastIoLock = FILEHIDEFastIoLock;
	fastIoDispatch->FastIoUnlockSingle = FILEHIDEFastIoUnlockSingle;
	fastIoDispatch->FastIoUnlockAll = FILEHIDEFastIoUnlockAll;
	fastIoDispatch->FastIoUnlockAllByKey = FILEHIDEFastIoUnlockAllByKey;
	fastIoDispatch->FastIoDeviceControl = FILEHIDEFastIoDeviceControl;
	fastIoDispatch->FastIoDetachDevice = FILEHIDEFastIoDetachDevice;
	fastIoDispatch->FastIoQueryNetworkOpenInfo = FILEHIDEFastIoQueryNetworkOpenInfo;
	fastIoDispatch->MdlRead = FILEHIDEFastIoMdlRead;
	fastIoDispatch->MdlReadComplete = FILEHIDEFastIoMdlReadComplete;
	fastIoDispatch->PrepareMdlWrite = FILEHIDEFastIoPrepareMdlWrite;
	fastIoDispatch->MdlWriteComplete = FILEHIDEFastIoMdlWriteComplete;
	fastIoDispatch->FastIoReadCompressed = FILEHIDEFastIoReadCompressed;
	fastIoDispatch->FastIoWriteCompressed = FILEHIDEFastIoWriteCompressed;
	fastIoDispatch->MdlReadCompleteCompressed = FILEHIDEFastIoMdlReadCompleteCompressed;
	fastIoDispatch->MdlWriteCompleteCompressed = FILEHIDEFastIoMdlWriteCompleteCompressed;
	fastIoDispatch->FastIoQueryOpen = FILEHIDEFastIoQueryOpen;

	DriverObject->FastIoDispatch = fastIoDispatch;

	status = ExInitializeResourceLite( &LockHash );
	if( !NT_SUCCESS( status ) )
	{
		ExFreePool (fastIoDispatch) ;
		DriverObject->FastIoDispatch = NULL ;
		IoDeleteDevice (g_DeviceObject) ;
		return status ;
	}

	for( i = 0; i < NUM_HASH; i++ )   
	{
		HashTable[ i ] = NULL;
	}

	g_DriverObject = DriverObject;

	InitializeListHead( &FsDeviceQueue );
	ExInitializeResource( &FsLock );

	deviceExtension = g_DeviceObject->DeviceExtension;
	deviceExtension->ullMagicNumber = MAGIC_NUMBER;
	deviceExtension->Type = FORAPP;
	deviceExtension->Drive = 0;

	RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );
	status = IoCreateSymbolicLink( &uniWin32NameString, &nameString );
	if (!NT_SUCCESS(status))
	{
		ExFreePool (fastIoDispatch) ;
		ExDeleteResource( &FsLock );
		IoDeleteDevice( g_DeviceObject );
		return status;
	}

	FILEHIDERecognizer();

	status = IoRegisterFsRegistrationChange( g_DriverObject, FILEHIDEFsNotification );
	if (!NT_SUCCESS( status )) 
	{
		IoUnregisterFsRegistrationChange( g_DriverObject, NULL );
		ExFreePool (fastIoDispatch) ;
		ExDeleteResource( &FsLock );
		IoDeleteDevice( g_DeviceObject );
		return status;
	}

	return status;
}

NTSTATUS
FILEHIDECommonCompletion(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    )
{
       PKEVENT pEvent = Context;

	KeSetEvent( pEvent, 0, FALSE );
	return STATUS_MORE_PROCESSING_REQUIRED;
}

NTSTATUS FILEHIDEPassThrough(IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp)
{
	PDEVICE_EXTENSION pDevExt;
	NTSTATUS ntStatus;
	PFILE_OBJECT pFileObj;
	WCHAR * pwFullPath = NULL;
	PIO_STACK_LOCATION pCurIrpStack;
	
	pDevExt = pDevObj->DeviceExtension;
	
	pCurIrpStack = IoGetCurrentIrpStackLocation(pIrp);
	
	pFileObj = pCurIrpStack->FileObject;
	
	if(pDevExt->ullMagicNumber == MAGIC_NUMBER)
	{
		if(pCurIrpStack->MajorFunction == IRP_MJ_CLOSE)
		{
				if( g_pEventObject != NULL )
				{
					ObDereferenceObject( g_pEventObject );
				}
				g_pEventObject = NULL;
				g_EventHandle = NULL;
				ntStatus = STATUS_SUCCESS;
		}
		pIrp->IoStatus.Status = ntStatus;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest( pIrp, IO_NO_INCREMENT );
		return ntStatus;
	}
	
	if(!fIsExistInHashTable(pFileObj, &pwFullPath))
	{
		return SkipAndSendIrpToNextDrive( pDevObj, pIrp );
	}
	
	switch(pCurIrpStack->MajorFunction)
	{
		case IRP_MJ_CLOSE:
			ExFreePool(pwFullPath);
			fFreeHashEntry(pFileObj);
			break;
		
		default:
			break;
	}
	
	return SkipAndSendIrpToNextDrive( pDevObj, pIrp );
}

NTSTATUS FILEHIDECreate(IN PDEVICE_OBJECT pDevObj, IN PIRP pIrp)
{
	PIO_STACK_LOCATION pCurIrpStack;
	NTSTATUS ntStatus;
	PDEVICE_EXTENSION pDevExt;
	WCHAR * pwFullPath = NULL;
	UCHAR * pFullPath = NULL;
	ULONG ulSize;
	ULONG preInformation;
	PFILE_OBJECT pFileObj;
	PHJ_NAME_INFORMATION pNameInfo;	
	ULONG dwRet;

	BOOLEAN blFolder;
		
	pDevExt = pDevObj->DeviceExtension;
	
	pCurIrpStack = IoGetCurrentIrpStackLocation(pIrp);
	pFileObj = pCurIrpStack->FileObject;
	
	if(pDevExt-> ullMagicNumber == MAGIC_NUMBER)
	{
		pIrp->IoStatus.Status = STATUS_SUCCESS;
		pIrp->IoStatus.Information = 0;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
		return STATUS_SUCCESS;
	}
	
	{ // FullPathϱ..
		ntStatus = HjCreateNameInformationMemory( &pNameInfo );
		if( ntStatus == STATUS_SUCCESS )
		{
			ntStatus = HjQueryNameString(
				pFileObj, 
				pNameInfo
				);
			if( ntStatus == STATUS_SUCCESS )
			{
				// FullPath  ̰ ´
			}
			else
			{	
				HjDeleteNameInformationMemory( pNameInfo );
				return SkipAndSendIrpToNextDrive( pDevObj, pIrp );
			}
		}
		else
		{
			return SkipAndSendIrpToNextDrive( pDevObj, pIrp );
		}
	}

	pwFullPath = (WCHAR *)HjAllocatePool(NonPagedPool, FILENAMESIZE * sizeof(WCHAR));
	if( !pwFullPath )
	{
		HjDeleteNameInformationMemory( pNameInfo );
		return SkipAndSendIrpToNextDrive( pDevObj, pIrp );
	}
	
	memcpy( (unsigned char *)pwFullPath, (unsigned char *)pNameInfo->uniName.Buffer, pNameInfo->uniName.Length );
	HjDeleteNameInformationMemory( pNameInfo );

	ntStatus = ReturnCurrentStackAfterSendIrpToNextDrive( pDevObj, pIrp);

	if( ntStatus == STATUS_SUCCESS )
	{
		fInsertHashEntry(pFileObj, pwFullPath );
	}
	else
	{
		ExFreePool( pwFullPath );
	}
	
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	return ntStatus;
}

