#include <ntddk.h>

#include "Queue.h"

/*  ----------------------------------------------------------------------------------------
	allocmem : ޸𸮸 Ҵϴ Լ
	----------------------------------------------------------------------------------------
	[Input]
	unsigned long dwSize : ҴϷ ޸ũ
	[Output]
	unsigned char * : Ҵ ޸ּ
    ---------------------------------------------------------------------------------------- */
unsigned char * allocmem( unsigned long dwSize )
{
	unsigned char * pTemp;
	pTemp = (unsigned char *)ExAllocatePool( NonPagedPool, dwSize );
	return pTemp;
}

/*  ----------------------------------------------------------------------------------------
	freemem : ޸𸮸 ϴ Լ
	----------------------------------------------------------------------------------------
	[Input]
	unsigned char * pBuffer : Ϸ ޸ּ
	[Output]
	none
    ---------------------------------------------------------------------------------------- */
void freemem( unsigned char * pBuffer )
{
	ExFreePool( pBuffer );
}

/*  ----------------------------------------------------------------------------------------
	Prolog, Epilog : CPU ϰų, ԽŰ Լ, 98迭 2000迭  ٸ
	----------------------------------------------------------------------------------------
	[Input]
	KIRQL :  Ų,  ϴ 켱  
	[Output]
	none
    ---------------------------------------------------------------------------------------- */
void Prolog( KIRQL * pIrql )
{
	KeRaiseIrql( DISPATCH_LEVEL, pIrql );
}

void Epilog( KIRQL Irql )
{
	KeLowerIrql( Irql );
}

/*  ----------------------------------------------------------------------------------------
	SizeofQueue : Queue  Ÿ ũ⸦ ϴ Լ
	----------------------------------------------------------------------------------------
	[Input]
	PHAJE_QUEUE pQueue :  Ǵ Queue
	[Output]
	unsigned long :  Ÿũ
    ---------------------------------------------------------------------------------------- */
unsigned long SizeofQueue( PHAJE_QUEUE pQueue )
{
	unsigned long dwSize;

	KIRQL Irql;
	Prolog( &Irql );

	if( pQueue->dwHead == pQueue->dwTail)
		dwSize = 0;
	else if( pQueue->dwHead < pQueue->dwTail)
		dwSize = pQueue->dwTail - pQueue->dwHead;
	else if( pQueue->dwHead > pQueue->dwTail)
		dwSize = MAXIMUMQUEUESIZE - pQueue->dwHead + pQueue->dwTail + 1;

	Epilog( Irql );
	return dwSize;
}

/*  ----------------------------------------------------------------------------------------
	AllocateQueue : Queue غϴ Լ
	----------------------------------------------------------------------------------------
	[Input]
	none
	[Output]
	PHAJE_QUEUE : غ Queue
    ---------------------------------------------------------------------------------------- */
PHAJE_QUEUE AllocateQueue()
{
	PHAJE_QUEUE pQueue;

	KIRQL Irql;
	Prolog( &Irql );

	pQueue = (PHAJE_QUEUE)allocmem( sizeof(HAJE_QUEUE) );

	if( pQueue )
	{
		pQueue->dwHead = pQueue->dwTail = 0;
	}
	
	Epilog( Irql );
	return pQueue;
}

/*  ----------------------------------------------------------------------------------------
	FreeQueue : Queue ϴ Լ
	----------------------------------------------------------------------------------------
	[Input]
	PHAJE_QUEUE pQueue : Ϸ Queue
	[Output]
	none
    ---------------------------------------------------------------------------------------- */
void FreeQueue( PHAJE_QUEUE pQueue )
{
	KIRQL Irql;
	Prolog( &Irql );

	freemem( (unsigned char *)pQueue );

	Epilog( Irql );
}

/*  ----------------------------------------------------------------------------------------
	InputQueue : Queue Ÿ ϴ Լ
	----------------------------------------------------------------------------------------
	[Input]
	PHAJE_QUEUE pQueue : Ϸ Queue
	unsigned char *DataBuffer : Ϸ Ÿ  ִ 
	unsigned long dwSize : Ϸ Ÿ Ÿ
	[Output]
	unsigned long :  Ÿ
    ---------------------------------------------------------------------------------------- */
unsigned long InputQueue( PHAJE_QUEUE pQueue, unsigned char *DataBuffer, unsigned long dwSize )
{
	unsigned long dwTemp, dwTemp2;

	KIRQL Irql;
	Prolog( &Irql );

	if( pQueue->dwTail < pQueue->dwHead ) // Queue ѹٲ..
	{
		dwTemp = pQueue->dwHead - pQueue->dwTail - 1;
		dwTemp = (dwTemp > dwSize )? dwSize:dwTemp;
		memcpy( &pQueue->Queue[pQueue->dwTail], DataBuffer, dwTemp );
		pQueue->dwTail += dwTemp;

		Epilog( Irql );
		return dwTemp;
	}

	dwTemp = MAXIMUMQUEUESIZE - pQueue->dwTail + 1;

	if( dwTemp > dwSize )
	{
		memcpy( &pQueue->Queue[pQueue->dwTail], DataBuffer, dwSize );
		pQueue->dwTail += dwSize;

		Epilog( Irql );
		return dwSize;
	}
	else
	{
		if( pQueue->dwHead == 0 )
		{
			memcpy( &pQueue->Queue[pQueue->dwTail], DataBuffer, dwTemp - 1 );
			pQueue->dwTail = MAXIMUMQUEUESIZE;

			Epilog( Irql );
			return dwTemp - 1;
		}
		else
		{
			memcpy( &pQueue->Queue[pQueue->dwTail], DataBuffer, dwTemp );
			DataBuffer += dwTemp;
			pQueue->dwTail = 0;
			
			dwTemp2 = pQueue->dwHead - 1;
			dwTemp2 = (dwTemp2 > (dwSize - dwTemp))? (dwSize - dwTemp):dwTemp2;

			memcpy( &pQueue->Queue[pQueue->dwTail], DataBuffer, dwTemp2 );

			pQueue->dwTail = dwTemp2;

			Epilog( Irql );
			return dwTemp2 + dwTemp;
		}
	}
}

/*  ----------------------------------------------------------------------------------------
	OutputQueue : Queueκ Ÿ  Լ
	----------------------------------------------------------------------------------------
	[Input]
	PHAJE_QUEUE pQueue : Ϸ Queue
	unsigned char *DataBuffer :  Ÿ Ϸ 
	unsigned long dwSize :  Ÿ Ÿ
	[Output]
	unsigned long :  Ÿ
    ---------------------------------------------------------------------------------------- */
unsigned long OutputQueue( PHAJE_QUEUE pQueue, unsigned char *DataBuffer, unsigned long dwSize )
{
	unsigned long dwTemp, dwTemp2;

	KIRQL Irql;
	Prolog( &Irql );

	if( pQueue->dwTail == pQueue->dwHead)
	{
		Epilog( Irql );
		return 0;
	}

	if( pQueue->dwTail > pQueue->dwHead ) 
	{
		dwTemp = pQueue->dwTail - pQueue->dwHead;
		dwTemp = (dwTemp > dwSize )? dwSize:dwTemp;
		memcpy( DataBuffer, &pQueue->Queue[pQueue->dwHead], dwTemp );
		pQueue->dwHead += dwTemp;

		Epilog( Irql );
		return dwTemp;
	}

	dwTemp = MAXIMUMQUEUESIZE - pQueue->dwHead + 1;

	if( dwTemp > dwSize )
	{
		memcpy( DataBuffer, &pQueue->Queue[pQueue->dwHead], dwSize );
		pQueue->dwHead += dwSize;

		Epilog( Irql );
		return dwSize;
	}
	else
	{
		memcpy( DataBuffer, &pQueue->Queue[pQueue->dwHead], dwTemp );
		DataBuffer += dwTemp;

		pQueue->dwHead = 0;
		
		dwTemp2 = pQueue->dwTail;
		dwTemp2 = (dwTemp2 > (dwSize - dwTemp))? (dwSize - dwTemp):dwTemp2;

		memcpy( DataBuffer, &pQueue->Queue[pQueue->dwHead], dwTemp2 );

		pQueue->dwHead = dwTemp2;

		Epilog( Irql );
		return dwTemp2 + dwTemp;
	}
}

void ResetQueue( PHAJE_QUEUE pQueue )
{
	KIRQL Irql;
	Prolog( &Irql );

	pQueue->dwHead = pQueue->dwTail = 0;

	Epilog( Irql );
}

/*  ----------------------------------------------------------------------------------------
	InputQueueWithoutNoPointerChange : Queue Ÿ ϴ Լ, ڷᱸͰ 
	----------------------------------------------------------------------------------------
	[Input]
	PHAJE_QUEUE pQueue : Ϸ Queue
	unsigned char *DataBuffer : Ϸ Ÿ  ִ 
	unsigned long dwSize : Ϸ Ÿ Ÿ
	[Output]
	unsigned long :  Ÿ
    ---------------------------------------------------------------------------------------- */
unsigned long InputQueueWithoutNoPointerChange( PHAJE_QUEUE pQueue, unsigned char *DataBuffer, unsigned long dwSize )
{
	unsigned long dwTemp, dwTemp2;

	unsigned long dwHead, dwTail;

	KIRQL Irql;
	Prolog( &Irql );

	dwHead = pQueue->dwHead;
	dwTail = pQueue->dwTail;

	if( dwTail < dwHead ) // Queue ѹٲ..
	{
		dwTemp = dwHead - dwTail - 1;
		dwTemp = (dwTemp > dwSize )? dwSize:dwTemp;
		memcpy( &pQueue->Queue[dwTail], DataBuffer, dwTemp );
		dwTail += dwTemp;

		Epilog( Irql );
		return dwTemp;
	}

	dwTemp = MAXIMUMQUEUESIZE - dwTail + 1;

	if( dwTemp > dwSize )
	{
		memcpy( &pQueue->Queue[dwTail], DataBuffer, dwSize );
		dwTail += dwSize;
		Epilog( Irql );
		return dwSize;
	}
	else
	{
		if( dwHead == 0 )
		{
			memcpy( &pQueue->Queue[dwTail], DataBuffer, dwTemp - 1 );
			dwTail = MAXIMUMQUEUESIZE;

			Epilog( Irql );
			return dwTemp - 1;
		}
		else
		{
			memcpy( &pQueue->Queue[dwTail], DataBuffer, dwTemp );
			DataBuffer += dwTemp;

			dwTail = 0;
			
			dwTemp2 = dwHead - 1;
			dwTemp2 = (dwTemp2 > (dwSize - dwTemp))? (dwSize - dwTemp):dwTemp2;

			memcpy( &pQueue->Queue[dwTail], DataBuffer, dwTemp2 );

			dwTail = dwTemp2;

			Epilog( Irql );
			return dwTemp2 + dwTemp;
		}
	}
}

/*  ----------------------------------------------------------------------------------------
	OutputQueueWithoutNoPointerChange : Queueκ Ÿ  Լ, ڷᱸͰ 
	----------------------------------------------------------------------------------------
	[Input]
	PHAJE_QUEUE pQueue : Ϸ Queue
	unsigned char *DataBuffer :  Ÿ Ϸ 
	unsigned long dwSize :  Ÿ Ÿ
	[Output]
	unsigned long :  Ÿ
    ---------------------------------------------------------------------------------------- */
unsigned long OutputQueueWithoutNoPointerChange( PHAJE_QUEUE pQueue, unsigned char *DataBuffer, unsigned long dwSize )
{
	unsigned long dwTemp, dwTemp2;

	unsigned long dwHead, dwTail;

	KIRQL Irql;
	Prolog( &Irql );

	dwHead = pQueue->dwHead;
	dwTail = pQueue->dwTail;

	if( dwTail > dwHead ) 
	{
		dwTemp = dwTail - dwHead;
		dwTemp = (dwTemp > dwSize )? dwSize:dwTemp;
		memcpy( DataBuffer, &pQueue->Queue[dwHead], dwTemp );
		dwHead += dwTemp;

		Epilog( Irql );
		return dwTemp;
	}

	dwTemp = MAXIMUMQUEUESIZE - dwHead + 1;

	if( dwTemp > dwSize )
	{
		memcpy( DataBuffer, &pQueue->Queue[dwHead], dwSize );
		dwHead += dwSize;

		Epilog( Irql );
		return dwSize;
	}
	else
	{
		memcpy( DataBuffer, &pQueue->Queue[dwHead], dwTemp );
		DataBuffer += dwTemp;

		dwHead = 0;
		
		dwTemp2 = dwTail;
		dwTemp2 = (dwTemp2 > (dwSize - dwTemp))? (dwSize - dwTemp):dwTemp2;

		memcpy( DataBuffer, &pQueue->Queue[dwHead], dwTemp2 );

		dwHead = dwTemp2;

		Epilog( Irql );
		return dwTemp2 + dwTemp;
	}
}

