#include <ntddk.h>

#include "VSerial.H"
#include "RS232.H"

void RS232_Emulation( PDEVICE_EXTENSION deviceExtension, ULONG dwfModemControlRegister )
{
	ULONG dwModemStatusRegister;

	dwModemStatusRegister = deviceExtension->MODEMSTATUSREGISTER;

	switch( dwfModemControlRegister )
	{
	case DTR:
		deviceExtension->MODEMCONTROLREGISTER |= DTR;
		if( !(deviceExtension->MODEMSTATUSREGISTER & DSR) )
		{
			dwModemStatusRegister |= DSR;
			dwModemStatusRegister |= DCD;
			dwModemStatusRegister |= TDCD;
			dwModemStatusRegister |= TDSR;
		}
		break;
	case ~DTR:
		deviceExtension->MODEMCONTROLREGISTER &= ~DTR;
		if( deviceExtension->MODEMSTATUSREGISTER & DSR )
		{
			dwModemStatusRegister &= ~DSR;
			dwModemStatusRegister &= ~DCD;
			dwModemStatusRegister |= TDCD;
			dwModemStatusRegister |= TDSR;
		}
		break;
	case RTS:
		deviceExtension->MODEMCONTROLREGISTER |= RTS;
		if( !(deviceExtension->MODEMSTATUSREGISTER & CTS) )
		{
			dwModemStatusRegister |= CTS;
			dwModemStatusRegister |= TCTS;
		}
		break;
	case ~RTS:
		deviceExtension->MODEMCONTROLREGISTER &= ~RTS;
		if( deviceExtension->MODEMSTATUSREGISTER & CTS )
		{
			dwModemStatusRegister &= ~CTS;
			dwModemStatusRegister |= TCTS;
		}
		break;
	}
	deviceExtension->MODEMSTATUSREGISTER = dwModemStatusRegister;
}

void pEmulationPollingFunction(
		IN PVOID context
	)
{
	PIRP PendingWaitIrp;
	PDEVICE_EXTENSION deviceExtension;
	PHYSICAL_ADDRESS dwInterval;
	NTSTATUS returnStatus;
	ULONG dwRet;
	KIRQL Irql;
	ULONG fChanged;

	KeSetPriorityThread( KeGetCurrentThread(), LOW_REALTIME_PRIORITY );
	deviceExtension = (PDEVICE_EXTENSION)context;

	dwInterval = RtlConvertLongToLargeInteger( -100000 );

	while( 1 )
	{
		returnStatus = KeWaitForSingleObject( &deviceExtension->EmulationPollingExitEvent, Executive, KernelMode, FALSE, &dwInterval );
		if( returnStatus == STATUS_SUCCESS )
		{
			KeSetEvent( &deviceExtension->EmulationPollingAnswerEvent, 0, FALSE );
			{
			    KeAcquireSpinLock(&deviceExtension->SpinLock, &Irql );
				while( PendingWaitIrp = OutputPendingWaitIrp( deviceExtension ) )
				{
					PendingWaitIrp->IoStatus.Status = STATUS_SUCCESS;
					PendingWaitIrp->IoStatus.Information = sizeof(ULONG);
					IoSetCancelRoutine( PendingWaitIrp, 0 );
					IoCompleteRequest ( PendingWaitIrp, IO_NO_INCREMENT );
				}
			    KeReleaseSpinLock(&deviceExtension->SpinLock, Irql );
			}
			return;
		}

	       KeAcquireSpinLock(&deviceExtension->SpinLock, &Irql );

		fChanged = deviceExtension->MODEMSTATUSREGISTER & (TDCD|TCTS|TDSR);
		PendingWaitIrp = OutputPendingWaitIrp( deviceExtension );
		if( !PendingWaitIrp )
		{
		       KeReleaseSpinLock(&deviceExtension->SpinLock, Irql );
			continue;
		}

		*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) = 0;

		if( (deviceExtension->Wm & SERIAL_EV_RXCHAR) && (SizeofQueue( deviceExtension->pHajeQueue )))
		{
			*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) |= SERIAL_EV_RXCHAR;
		}

		if( !fChanged && !(*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer)))
		{
			InputPendingWaitIrp(deviceExtension, PendingWaitIrp );
		       KeReleaseSpinLock(&deviceExtension->SpinLock, Irql );
			continue;
		}

		if( (deviceExtension->Wm & SERIAL_EV_RLSD) && (fChanged & TDCD) )
			*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) |= SERIAL_EV_RLSD;
		if( (deviceExtension->Wm & SERIAL_EV_CTS) && (fChanged & TCTS) )
			*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) |= SERIAL_EV_CTS;
		if( (deviceExtension->Wm & SERIAL_EV_DSR) && (fChanged & TDSR) )
			*(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) |= SERIAL_EV_DSR;

		if( *(ULONG *)(PendingWaitIrp->AssociatedIrp.SystemBuffer) )
		{
			PendingWaitIrp->IoStatus.Status = STATUS_SUCCESS;
			PendingWaitIrp->IoStatus.Information = sizeof(ULONG);
			IoSetCancelRoutine( PendingWaitIrp, 0 );
			IoCompleteRequest ( PendingWaitIrp, IO_NO_INCREMENT );
		}
		else
		{
			InputPendingWaitIrp(deviceExtension, PendingWaitIrp );
		}
	       KeReleaseSpinLock(&deviceExtension->SpinLock, Irql );
	}
}


